﻿using System.Net;

namespace UlstuGosApi;

public static class TimeTableRouteGroup
{
    public static RouteGroupBuilder AddTimeTable(this RouteGroupBuilder group)
    {
        group.MapGet("/groups", () => TypedResults.Ok(groups))
            .WithSummary("Список групп");
        group.MapGet("/rooms", () => TypedResults.Ok(rooms))
            .WithSummary("Список аудиторий");
        group.MapGet("/teachers", () => TypedResults.Ok(teachers))
            .WithSummary("Список преподавателей");

        group.MapGet("/by-group/{groupName}", (string groupName) =>
        {
            var group = groups.FirstOrDefault(m => m.Name == groupName);
            if (group == default)
            {
                return Results.NotFound();
            }
            return Results.Ok(GenerateByGroup(group));
        })
            .Produces<TimetableWeek[]>()
            .WithSummary("Расписание по учебной группе");

        group.MapGet("/by-room/{roomName}", (string roomName) =>
        {
            var room = rooms.FirstOrDefault(m => m.Name == WebUtility.UrlDecode(roomName));
            if (room == default)
            {
                return Results.NotFound();
            }
            return Results.Ok(GenerateByRoom(room));

        })
            .Produces<TimetableWeek[]>()
            .WithSummary("Расписание по аудитории");

        group.MapGet("/by-teacher/{teacherFullName}", (string teacherFullName) =>
        {
            var teacher = teachers.FirstOrDefault(m => m.FullName == teacherFullName);
            if (teacher == default)
            {
                return Results.NotFound();
            }
            return Results.Ok(GenerateByTeacher(teacher));
        })
            .Produces<TimetableWeek[]>()
            .WithSummary("Расписание по преподавателю");

        return group;
    }

    private static TimetableGroup[] groups = [
        new TimetableGroup("ПИбд-41"),
        new TimetableGroup("ПИбд-42"),
        new TimetableGroup("ПИбд-43"),
        ];

    private static TimetableRoom[] rooms = [
        new TimetableRoom("3/420"),
        new TimetableRoom("3/424"),
        new TimetableRoom("3/429"),
        new TimetableRoom("3/431"),
        ];

    private static TimetableTeacher[] teachers = [
        new TimetableTeacher("Иванов И.И."),
        new TimetableTeacher("Петров П.П."),
        new TimetableTeacher("Сидорова С.С."),
        ];

    private static TimetableSubject[] subjects = [
        new TimetableSubject(teachers[0], "Основы программирования", "лек"),
        new TimetableSubject(teachers[0], "Базы данных", "лек"),
        new TimetableSubject(teachers[0], "Схемотехника", "лек"),
        new TimetableSubject(teachers[1], "Основы программирования", "л/р"),
        new TimetableSubject(teachers[1], "Базы данных", "л/р"),
        new TimetableSubject(teachers[2], "Философия", "лек"),
        new TimetableSubject(teachers[2], "Философия", "сем"),
        ];

    private const int LessonsMaxCount = 2 * 6 * 8;
    private static List<TimetableLesson> lessons = new();
    private static object lessonsLockObj = new object();
    public static IReadOnlyList<TimetableLesson> Lessons
    {
        get
        {
            lock (lessonsLockObj)
            {
                if (lessons.Count == 0)
                {
                    GenerateLessons();
                }
            }
            return lessons;
        }
    }
    private static void GenerateLessons()
    {
        foreach (var idx in Enumerable.Range(0, LessonsMaxCount))
        {
            lessons.Add(
                new TimetableLesson(
                    groups[Random.Shared.Next(groups.Length)], rooms[Random.Shared.Next(rooms.Length)], subjects[Random.Shared.Next(subjects.Length)]
                ));
        }
    }

    private static TimetableWeek[] GenerateByGroup(TimetableGroup group) => GenerateWithLessons(Lessons.Where(i => i.Group == group));
    private static TimetableWeek[] GenerateByRoom(TimetableRoom room) => GenerateWithLessons(Lessons.Where(i => i.Room == room));
    private static TimetableWeek[] GenerateByTeacher(TimetableTeacher teacher) => GenerateWithLessons(Lessons.Where(i => i.Subject.Teacher == teacher));
    private static TimetableWeek[] GenerateWithLessons(IEnumerable<TimetableLesson> filteredLessons)
    {
        var weeks = new List<TimetableWeek>();
        var lessonsEnumerator = filteredLessons.GetEnumerator();
        weeks.Add(new TimetableWeek(GenerateDays(lessonsEnumerator), true));
        weeks.Add(new TimetableWeek(GenerateDays(lessonsEnumerator), false));
        return weeks.ToArray();
    }
    private static TimetableDay[] GenerateDays(IEnumerator<TimetableLesson> filteredLessons)
    {
        var result = new List<TimetableDay>();
        foreach (var _ in Enumerable.Range(1, 6))
        {
            result.Add(GenerateDay(filteredLessons));
        }
        return result.ToArray();
    }
    private static TimetableDay GenerateDay(IEnumerator<TimetableLesson> filteredLessons)
    {
        var result = new List<TimetableLesson?>();
        foreach (var _ in Enumerable.Range(1, 8))
        {
            if (Random.Shared.NextDouble() < 0.25)
            {
                filteredLessons.MoveNext();
                result.Add(filteredLessons.Current);
            }
            else
            {
                result.Add(null);
            }
        }
        return new TimetableDay(result.ToArray());
    }
}

/// <summary>
/// Неделя расписания.
/// </summary>
/// <param name="Days">Массив дней недели: ПН, ВТ, СР и до субботы.</param>
/// <param name="IsOdd">Флаг нечётности.</param>
public sealed record TimetableWeek(
    TimetableDay[] Days,
    bool IsOdd
    );

/// <summary>
/// День расписания.
/// </summary>
/// <param name="Lessons">Массив пар от 1й до 8й. Если пары в "слоте" нет, возвращается null.</param>
public sealed record TimetableDay(
    TimetableLesson?[] Lessons
    );

/// <summary>
/// "Слот" расписания, непосредственно пара.
/// </summary>
/// <param name="Group">Группа.</param>
/// <param name="Room">Аудитория.</param>
/// <param name="Subject">Дисциплина.</param>
public sealed record TimetableLesson(
    TimetableGroup Group,
    TimetableRoom Room,
    TimetableSubject Subject
    );

/// <summary>
/// Группа.
/// </summary>
/// <param name="Name">Название.</param>
public sealed record TimetableGroup(
    string Name
    );

/// <summary>
/// Аудитория.
/// </summary>
/// <param name="Name">Номер.</param>
public sealed record TimetableRoom(
    string Name
    );

/// <summary>
/// Дисциплина.
/// </summary>
/// <param name="Teacher">Преподаватель.</param>
/// <param name="Name">Имя.</param>
/// <param name="Type">Тип занятия.</param>
public sealed record TimetableSubject(
    TimetableTeacher Teacher,
    string Name,
    string Type
    );

/// <summary>
/// Преподаватель
/// </summary>
/// <param name="FullName">ФИО.</param>
public sealed record TimetableTeacher(
    string FullName
    );
